--              Syntax Color
-- These routines are used to display Euphoria program lines
-- in multiple colors. The editor (ed.ex) and the pretty printer (eprint.ex)
-- both include this file.

-- inspired by ex12.exw of David Cuny's win32lib
-- modified from euphoria\bin\syncolor.e

-- Euphoria keywords and routines built in to the interpreter, ex.exe

global constant keywords = {
    "if", "end", "then", "procedure", "else", "for", "return",
    "do", "elsif", "while", "type", "constant", "to", "and", "or",
    "exit", "function", "global", "by", "not", "include",
    "with", "without"}

global constant builtins = {
    "length", "puts", "integer", "sequence", "position", "object",
    "append", "prepend", "print", "printf", 
    "clear_screen", "floor", "getc", "gets", "get_key",
    "rand", "repeat", "atom", "compare", "find", "match",
    "time", "command_line", "open", "close", "trace", "getenv",
    "sqrt", "sin", "cos", "tan", "log", "system", "date", "remainder",
    "power", "machine_func", "machine_proc", "abort", "peek", "poke", 
    "call", "sprintf", "arctan", "and_bits", "or_bits", "xor_bits",
    "not_bits", "pixel", "get_pixel", "mem_copy", "mem_set",
    "c_proc", "c_func", "routine_id", "call_proc", "call_func", 
    "poke4", "peek4s", "peek4u"}

global constant library = {
    "all_palette", "allocate", "allocate_low", "allocate_string",
    "allow_break", "atom_to_float32", "atom_to_float64", "bits_to_int",
    "bk_color", "bytes_to_int", "call_back", "check_break", "crash_message",
    "current_dir", "cursor", "custom_sort", "define_c_func", "define_c_proc",
    "dir", "display_image", "display_text_image", "dos_interrupt",
    "draw_line", "ellipse", "float32_to_atom", "float64_to_atom",
    "free", "free_low", "get", "get_active_page", "get_all_palette",
    "get_display_page", "get_mouse", "get_position", "get_vector",
    "graphics_mode", "int_to_bits", "int_to_bytes", "lock_memory",
    "lower", "message_box", "mouse_events", "mouse_pointer",
    "open_dll", "palette", "polygon", "read_bitmap", "save_bitmap",
    "save_image", "save_screen", "save_text_image", "scroll",
    "seek", "set_active_page", "set_display_page", "set_rand",
    "set_vector", "sort", "sound", "text_rows", "tick_rate",
    "upper", "use_vesa", "value", "video_config", "wait_key",
    "where", "wildcard_file", "wildcard_match", "wrap"}


constant             -- colors of various syntax classes
    Black = #000000,
    Gray = #AAAAAA,
    Green = #00AA00,
    Yellow = #00AAAA,
    Magenta = #AA00AA,
    Cyan = #AAAA00,
    Red = #0000AA,
    Blue = #AA0000

-- Colors for syntax elements
global constant NORMAL_COLOR = Black,
                COMMENT_COLOR = Cyan,
                KEYWORD_COLOR = Green,
                BUILTIN_COLOR = Magenta,
                LIBRARY_COLOR = Red,
                STRING_COLOR = Blue,
                BRACKET_COLOR = {Black, Yellow, Magenta, Cyan, Green}


-- character classes
constant DIGIT = 1,
         OTHER = 2,
         LETTER  = 3,
         BRACKET = 4,
         QUOTE   = 5,
         DASH = 6,
         WHITE_SPACE = 7,
         NEW_LINE = 8

sequence char_class

global procedure init_class()
-- set up character classes for easier line scanning
-- (assume no 0 char)
    char_class = repeat(OTHER, 255)

    char_class['a'..'z'] = LETTER
    char_class['A'..'Z'] = LETTER
    char_class['_'] = LETTER
    char_class['0'..'9'] = DIGIT
    char_class['['] = BRACKET
    char_class[']'] = BRACKET
    char_class['('] = BRACKET
    char_class[')'] = BRACKET
    char_class['{'] = BRACKET
    char_class['}'] = BRACKET
    char_class['\''] = QUOTE
    char_class['"'] = QUOTE
    char_class[' '] = WHITE_SPACE
    char_class['\t'] = WHITE_SPACE
--    char_class['\n'] = NEW_LINE
    char_class['\n'] = WHITE_SPACE
    char_class['-'] = DASH
end procedure

sequence line  -- the line being processed
integer seg_start, seg_end -- start and end of current segment of line
integer color  -- the current color

constant CFM_BOLD  =      1
constant CFM_COLOR =       1073741824
constant CFM_FACE   =     536870912
constant CFM_ITALIC  =      2
constant CFM_OFFSET   =     268435456
constant CFM_PROTECTED =       16
constant CFM_SIZE      =  #80000000
constant CFM_STRIKEOUT =       8
constant CFM_UNDERLINE =       4
--constant EM_SETCHARFORMAT = 1092
constant SCF_SELECTION = 1

constant CFE_AUTOCOLOR =       1073741824
constant CFE_BOLD      =  1
constant CFE_ITALIC    =    2
constant CFE_STRIKEOUT =       8
constant CFE_UNDERLINE =       4
constant CFE_PROTECTED =       16


constant charformat = allocate(60)
mem_set(charformat, 0, 60)
poke4(charformat, 60)      -- picked 60 by trial and error
poke4(charformat+4, CFM_COLOR)
--poke4(charformat+CHARFORMAT_dwEffects, CFE_BOLD)

atom hedit, junk

procedure flush(integer new_color)
-- if the color is changing, write out the current segment
    if new_color != color then
        if color != -1 then
            poke4(charformat+20, color)
            junk = c_func(SendMessage, {hedit, EM_SETSEL, seg_start-1, seg_end})
            junk = c_func(SendMessage, {hedit, EM_SETCHARFORMAT, SCF_SELECTION, charformat})
            if not junk then
                trace(1)
            end if
--            text_color(color)
--            puts(SCREEN, line[seg_start..seg_end])
            seg_start = seg_end + 1
        end if
        color = new_color
    end if
end procedure

global procedure BuildColorText(atom hwnd, sequence text)
-- Display a '\n'-terminated line with colors identifying the various
-- parts of the Euphoria language.
-- Each screen write has a lot of overhead, so we try to minimize
-- the number of them by collecting consecutive characters of the
-- same color into a 'segment' seg_start..seg_end.
    integer class, last, i, c, bracket_level
    sequence word

    hedit = hwnd
    line = text
    color = -1 -- initially undefined
    bracket_level = 0
    seg_start = 1
    seg_end = 0

    while seg_end < length(line) do
        c = line[seg_end+1]
        class = char_class[c]

        if class = WHITE_SPACE then
            seg_end = seg_end + 1 -- continue with same color

        elsif class = LETTER then
            last = length(line)-1
            for j = seg_end + 2 to last do
                c = line[j]
                class = char_class[c]
                if class != LETTER then
                    if class != DIGIT then
                        last = j - 1
                        exit
                    end if
                end if
            end for
            word = line[seg_end+1..last]
            if find(word, keywords) then
                flush(KEYWORD_COLOR)
            elsif find(word, builtins) then
                flush(BUILTIN_COLOR)
            elsif find(word, library) then
                flush(LIBRARY_COLOR)
            else
                flush(NORMAL_COLOR)
            end if
            seg_end = last

        elsif class <= OTHER then -- DIGIT too
            flush(NORMAL_COLOR)
            seg_end = seg_end + 1

        elsif class = BRACKET then
            if find(c, "([{") then
                bracket_level = bracket_level + 1
            end if
            if bracket_level >= 1 and
               bracket_level <= length(BRACKET_COLOR) then
                flush(BRACKET_COLOR[bracket_level])
            else
                flush(NORMAL_COLOR)
            end if
            if find(c, ")]}") then
                bracket_level = bracket_level - 1
            end if
            seg_end = seg_end + 1

-- NEW_LINE is treated as WHITE_SPACE
--        elsif class = NEW_LINE then  
--            exit  -- end of line
        
        elsif class = DASH then
            if line[seg_end+2] = '-' then
                flush(COMMENT_COLOR)
                seg_end = seg_end + 3
                while line[seg_end] != '\n' do
                    seg_end = seg_end + 1
                end while
--                seg_end = length(line)-1
            else
                flush(NORMAL_COLOR)
                seg_end = seg_end + 1
            end if

        else  -- QUOTE
            i = seg_end + 2
            while i <= length(line) do
                if line[i] = c then
                    i = i + 1
                    exit
                elsif line[i] = '\\' then
                    if i < length(line)-1 then
                        i = i + 1 -- ignore escaped char
                    end if
                elsif line[i] = '\n' then
                    exit
                end if
                i = i + 1
            end while
            flush(STRING_COLOR)
            seg_end = i - 1
        end if
    end while

    if color != -1 then
--        text_color(color)
            poke4(charformat+20,color)
    end if
--    puts(SCREEN, line[seg_start..seg_end])
    junk = c_func(SendMessage, {hedit, EM_SETSEL, seg_start-1, seg_end-1})
    junk = c_func(SendMessage, {hedit, EM_SETCHARFORMAT, SCF_SELECTION, charformat})
    junk = c_func(SendMessage, {hedit, EM_SETSEL, 0, 0})
end procedure


